
#include "fggl/gfx/ogl/types.hpp"
#include <fggl/gfx/ogl/shader.hpp>

#include <iostream>
#include <vector>
#include <spdlog/spdlog.h>

using namespace fggl::gfx;

bool ShaderCache::compileShader(const std::string& fname, GLuint sid) {
	std::string source;
	bool result = m_storage->load(fggl::data::Data, fname, &source);
	if ( !result ) {
        spdlog::warn("could not load shader source from disk: {}", fname.c_str());
		return false;
	}

	// upload and compile shader
	const auto *src_cstr = (const GLchar *)source.c_str();
	glShaderSource(sid, 1, &src_cstr, 0);
	glCompileShader(sid);

	// check it worked
	GLint compiled = GL_FALSE;
	glGetShaderiv(sid, GL_COMPILE_STATUS, &compiled);
	if ( compiled == GL_FALSE ) {

		GLint maxLength = 0;
		glGetShaderiv(sid, GL_INFO_LOG_LENGTH, &maxLength);
		char* infoLog = new char[maxLength];
		glGetShaderInfoLog(sid, maxLength, &maxLength, infoLog);

        spdlog::warn("could not compile shader, '{}': {}", fname, infoLog);
		delete[] infoLog;

		return false;
	}

	return true;
}

ShaderCache::ShaderCache(std::shared_ptr<fggl::data::Storage> storage) : m_storage(storage), m_shaders(), m_binary(true) {

	if ( !GLAD_GL_ARB_get_program_binary ) {
		spdlog::warn("the graphics card doesn support shader caching, disabling");
		m_binary = false;
	}

}

bool ShaderCache::loadFromDisk(GLuint pid, const ShaderConfig& config) {

	BinaryCache cache;
	auto fname = "shader_" + config.name + ".bin";
	bool status = m_storage->load( fggl::data::Cache, fname, &cache );

	if ( !status ) {
        spdlog::info("cached shader '{}' could not be loaded from disk", config.name);
		return false;
	}

	bool result = cacheLoad(pid, &cache);
	std::free(cache.data);
	return result;
}

void ShaderCache::saveToDisk(GLuint pid, const ShaderConfig& config) {

	BinaryCache cache;
	cacheSave(pid, &cache);

	auto fname = "shader_" + config.name + ".bin";
	m_storage->save( fggl::data::Cache, fname, &cache);
}

GLuint ShaderCache::getOrLoad(const ShaderConfig& config) {
	try {
		return m_shaders.at(config.name);
	} catch ( std::out_of_range& e) {
		return load(config);
	}
}

GLuint ShaderCache::get(const std::string& name) {
	return m_shaders.at(name);
}

GLuint ShaderCache::load(const ShaderConfig& config) {

    spdlog::debug("starting shader program generation for {}", config.name);

	GLuint pid = glCreateProgram();

	if ( m_binary ) {
		// if we have support for shader cache, give that a go
		bool worked = loadFromDisk(pid, config);
		if ( worked ) {
			m_shaders[config.name] = pid;
			return pid;
		}

        spdlog::debug("could not use cached shader for '{}', doing full compile.", config.name);
	}

	// TODO actual shader loading
	GLuint vertShader = glCreateShader(GL_VERTEX_SHADER);
	compileShader(config.vertex, vertShader);
	glAttachShader(pid, vertShader);

	GLuint fragShader = glCreateShader(GL_FRAGMENT_SHADER);
	compileShader(config.fragment, fragShader);
	glAttachShader(pid, fragShader);

	GLuint geomShader;
	if ( config.hasGeom ) {
		geomShader = glCreateShader(GL_GEOMETRY_SHADER);
		compileShader( config.geometry, geomShader );
		glAttachShader(pid, geomShader);
	}

	glLinkProgram(pid);
	glDetachShader(pid, vertShader);
	glDetachShader(pid, fragShader);

	if ( config.hasGeom ) {
		glDetachShader(pid, geomShader);
	}

	GLint linked = GL_FALSE;
	glGetProgramiv(pid, GL_LINK_STATUS, &linked);
	if ( linked == GL_FALSE ) {

		// get the error
		std::array<char, 512> infoLog;
		glGetProgramInfoLog(pid, infoLog.size(), nullptr, infoLog.data());
        spdlog::warn("linking shader program '{}' failed: {}", config.name, infoLog.data());

		// cleanup
		glDeleteProgram( pid );
		glDeleteShader( vertShader );
		glDeleteShader( fragShader );
		if ( config.hasGeom ) {
			glDeleteShader( geomShader );
		}

		return false;
	}
		
	if ( m_binary ) {
		saveToDisk(pid, config);
	}

	// update the cache and return
	m_shaders[ config.name ] = pid;
	return pid;
}

void ShaderCache::cacheSave(GLuint pid, BinaryCache* cache) {
	GLsizei length;
	glGetProgramiv(pid, GL_PROGRAM_BINARY_LENGTH, &length);

	cache->data = std::malloc(length);
	cache->size = length;

	glGetProgramBinary(pid, length, &cache->size, &cache->format, cache->data);
}

bool ShaderCache::cacheLoad(GLuint pid, const BinaryCache* cache) {
	if ( !m_binary ) {
		return false;
	}
	glProgramBinary(pid, cache->format, cache->data, cache->size);

	// check it loaded correctly
	GLint status = GL_FALSE;
	glGetProgramiv(pid, GL_LINK_STATUS, &status);
	return status == GL_TRUE;
}

template<>
bool fggl::data::fggl_deserialize(std::filesystem::path &data, fggl::gfx::BinaryCache *out) {
	auto f = 
	#ifdef _MSC_VER
		_wfopen(data.c_str(), L"r");
	#else
		std::fopen( data.c_str(), "r");
	#endif

	if ( f == nullptr ) {
        spdlog::warn("could not load cached shader, fp was null");
		return false;
	}

	auto rsize = std::fread( &out->format, sizeof(GLenum), 1, f);
	if ( rsize != 1 ) {
        spdlog::warn("could not load cached shader: type read failed");
		std::fclose(f);
		return false;
	}

	out->size = 0;
	rsize = std::fread( &out->size, sizeof(GLsizei), 1, f);
	if ( rsize != 1 ) {
        spdlog::warn("could not load cached shader: size read failed");
		std::fclose(f);
		return false;
	}

	out->data = std::malloc(out->size);
	auto readSize = std::fread( out->data, out->size, 1, f );

	auto result = true;
	if ( readSize != 1 ) {
        spdlog::warn("could not load cached shader: reading failed!");
		std::free( out->data );
		result = false;
	}

	std::fclose(f);
	return result;
}

#include <iostream>
#include <fstream>

template<>
bool fggl::data::fggl_deserialize(std::filesystem::path& data, std::string *out) {
	std::ifstream ifs(data);
	out->assign( (std::istreambuf_iterator<char>(ifs)),
		     (std::istreambuf_iterator<char>()) );
	return true;
}



template<>
bool fggl::data::fggl_serialize(std::filesystem::path &data, const fggl::gfx::BinaryCache *out) {

	// try and write
	auto f =
		#ifdef _MSC_VER
			_wfopen( data.c_str(), L"w");
		#else
			std::fopen( data.c_str(), "w");
		#endif

	if ( f == nullptr ){
		return false;
	}

	std::fwrite( &out->format, sizeof(GLenum), 1, f);
	std::fwrite( &out->size, sizeof(GLsizei), 1, f);
	std::fwrite( out->data, out->size, 1, f);

	std::fclose(f);
	std::free( out->data );

	return true;
}
