//
// Created by webpigeon on 20/11/2021.
//

#include <fggl/ecs3/ecs.hpp>
#include <fggl/input/input.hpp>

#include <fggl/gfx/camera.hpp>
#include <fggl/input/camera_input.h>

namespace fggl::input {

    void process_arcball(fggl::ecs3::World &ecs, const Input &input, fggl::ecs::entity_t cam) {
        // see https://asliceofrendering.com/camera/2019/11/30/ArcballCamera/
        auto *camTransform = ecs.get<fggl::math::Transform>(cam);
        auto *camComp = ecs.get<fggl::gfx::Camera>(cam);
        auto &mouse = input.mouse;

        glm::vec4 position(camTransform->origin(), 1.0f);
        glm::vec4 pivot(camComp->target, 1.0f);
        glm::mat4 view = glm::lookAt(camTransform->origin(), camComp->target, camTransform->up());
        glm::vec3 viewDir = -glm::transpose(view)[2];
        glm::vec3 rightDir = glm::transpose(view)[0];

        float deltaAngleX = (2 * M_PI);
        float deltaAngleY = (M_PI);
        float xAngle = (-mouse.axisDelta(fggl::input::MouseAxis::X)) * deltaAngleX;
        float yAngle = (-mouse.axisDelta(fggl::input::MouseAxis::Y)) * deltaAngleY;

        // rotate the camera around the pivot on the first axis
        glm::mat4x4 rotationMatrixX(1.0f);
        rotationMatrixX = glm::rotate(rotationMatrixX, xAngle, fggl::math::UP);
        position = (rotationMatrixX * (position - pivot)) + pivot;

        // rotate the camera aroud the pivot on the second axis
        glm::mat4x4 rotationMatrixY(1.0f);
        rotationMatrixY = glm::rotate(rotationMatrixY, yAngle, rightDir);
        glm::vec3 finalPos = (rotationMatrixY * (position - pivot)) + pivot;

        camTransform->origin(finalPos);
    }

    void process_freecam(fggl::ecs3::World &ecs, const Input &input, fggl::ecs::entity_t cam) {
        float rotationValue = 0.0f;
        glm::vec3 translation(0.0f);

        auto &keyboard = input.keyboard;
        auto* settings = ecs.get<FreeCamKeys>(cam);

        // calculate rotation (user input)
        if (keyboard.down(settings->rotate_cw)) {
            rotationValue = ROT_SPEED;
        } else if (keyboard.down(settings->rotate_ccw)) {
            rotationValue = -ROT_SPEED;
        }

        // calculate movement (user input)
        if (keyboard.down(settings->forward)) {
            translation -= fggl::math::RIGHT;
        }

        if (keyboard.down(settings->backward)) {
            translation += fggl::math::RIGHT;
        }

        if (keyboard.down(settings->right)) {
            translation += fggl::math::FORWARD;
        }

        if (keyboard.down(settings->left)) {
            translation -= fggl::math::FORWARD;
        }

        // apply rotation/movement
        auto camTransform = ecs.get<fggl::math::Transform>(cam);
        auto camComp = ecs.get<fggl::gfx::Camera>(cam);

        glm::vec4 position(camTransform->origin(), 1.0f);
        glm::vec4 pivot(camComp->target, 1.0f);

        // apply movement
        if (translation != glm::vec3(0.0f)) {
            const auto rotation = (position - pivot);
            const float angle = atan2f(rotation.x, rotation.z);
            const auto rotationMat = glm::rotate(MAT_IDENTITY, angle, fggl::math::UP);

            auto deltaMove = (rotationMat * glm::vec4(translation, 1.0f)) * PAN_SPEED;
            deltaMove.w = 0.0f;

            position += deltaMove;
            pivot += deltaMove;
        }

        // apply rotation
        if (rotationValue != 0.0f) {
            glm::mat4 rotation = glm::rotate(MAT_IDENTITY, rotationValue, fggl::math::UP);
            position = (rotation * (position - pivot)) + pivot;
        }

        camTransform->origin(position);
        camComp->target = pivot;
    }

    void process_edgescroll(fggl::ecs3::World &ecs, const Input &input, fggl::ecs::entity_t cam) {
        glm::vec3 translation(0.0f);

        auto &mouse = input.mouse;

        // calculate movement (user input)
        if (mouse.axis(MouseAxis::Y) < 0.9f) {
            translation -= fggl::math::RIGHT;
        }

        if (mouse.axis(MouseAxis::Y) > -0.9f) {
            translation += fggl::math::RIGHT;
        }

        if (mouse.axis(MouseAxis::X) > -0.9f) {
            translation += fggl::math::FORWARD;
        }

        if (mouse.axis(MouseAxis::X) < 0.9f) {
            translation -= fggl::math::FORWARD;
        }

        // apply rotation/movement
        auto camTransform = ecs.get<fggl::math::Transform>(cam);
        auto camComp = ecs.get<fggl::gfx::Camera>(cam);

        glm::vec4 position(camTransform->origin(), 1.0f);
        glm::vec4 pivot(camComp->target, 1.0f);

        // apply movement
        if (translation != glm::vec3(0.0f)) {
            const auto rotation = (position - pivot);
            const float angle = atan2f(rotation.x, rotation.z);
            const auto rotationMat = glm::rotate(MAT_IDENTITY, angle, fggl::math::UP);

            auto deltaMove = (rotationMat * glm::vec4(translation, 1.0f)) * PAN_SPEED;
            deltaMove.w = 0.0f;

            position += deltaMove;
            pivot += deltaMove;
        }

        // move camera
        camTransform->origin(position);
        camComp->target = pivot;
    }
}
