#ifndef FGGL_ECS_H
#define FGGL_ECS_H

#include "utility.hpp"
#include "component.hpp"

#include <iostream>
#include <algorithm>
#include <cassert>
#include <string>
#include <vector>
#include <unordered_map>

namespace fggl::ecs {
	using archToken_t = std::vector<component_type_t>;

	struct Archetype {
		constexpr static unsigned int default_cap = 0;
		const archToken_t type;
		std::vector<ComponentBase::data_t *> data;
		std::vector<std::size_t> dataSizes;
		std::vector<entity_t> entities;

		Archetype(const archToken_t &type_a);

		inline archToken_t create(const component_type_t cid) const {
			assert(!contains(cid));

			// create the new type
			auto newType = type;
			newType.push_back(cid);
			std::sort(newType.begin(), newType.end());
			return newType;
		}

		inline bool contains(const component_type_t cid) const {
			return (std::find(type.begin(), type.end(), cid) != type.end());
		}
	};

	class ECS {
			struct Record {
				Archetype *archetype;
				std::size_t index;
			};
			using componentmap_t = std::unordered_map<component_type_t, ComponentBase *>;
			using entitymap_t = std::unordered_map<entity_t, Record>;
			using archetype_t = std::vector<Archetype *>;

		public:
			ECS();
			~ECS();

			entity_t getNewID();
			entity_t createEntity();
			void removeEntity(const entity_t eid);

			template<class C>
			void registerComponent() {
				component_type_t type = Component<C>::typeID();
				if (m_componentMap.find(type) != m_componentMap.end())
					return;
				m_componentMap.emplace(type, new Component<C>);
			}

			template<class C>
			bool isComponentRegistered() {
				component_type_t type = Component<C>::typeID();
				return (m_componentMap.find(type) != m_componentMap.end());
			}

			template<class C, typename... Args>
			C *addComponent(const entity_t &id, Args &&... args) {
				component_type_t type = Component<C>::typeID();
				assert(isComponentRegistered<C>());

				Record &record = m_entityArchtypes[id];
				Archetype *oldArch = record.archetype;
				C *newComp = nullptr;
				Archetype *newArch = nullptr;

				if (!oldArch) {
					archToken_t newID(1, type);
					const ComponentBase *const newCompType = m_componentMap[type];

					// fetch type
					newArch = getArchetype(newID);
					assert(newArch->type.size() == 1);

					// calculate if we have enouph space to allocate
					std::size_t emplacementPos = ensureCapacity(newArch, 0, newCompType);
					newComp = new(&newArch->data[0][emplacementPos])C(std::forward<Args>(args)...);
				} else {
					// check if the arch contains the component
					if (oldArch->contains(type)) {
						return nullptr;
					}

					// create a new archetype with the component
					auto newID = oldArch->create(type);
					newArch = getArchetype(newID);

					// relocate the old data to the new archetype
					for (std::size_t j = 0; j < newID.size(); ++j) {
						const component_type_t compType = newID[j];
						const ComponentBase *const comp = m_componentMap.at(compType);


						// TODO this seems a little suspect - surely we could allocate all blocks at once?
						// if per component the arrays could become out of sync...
						int newOffset = ensureCapacity(newArch, j, comp);
						int oldIdx = getComponentIdx(oldArch, compType);

						if (oldIdx != -1) {
							assert(oldArch->contains(compType));
							const std::size_t compSize = comp->size();
							const std::size_t oldOffset = record.index * compSize;

							comp->move(&oldArch->data[oldIdx][oldOffset],
									   &newArch->data[j][newOffset]);
							comp->destroy(&oldArch->data[oldIdx][oldOffset]);
						} else {
							assert(!oldArch->contains(compType));
							newComp = new(&newArch->data[j][newOffset])
								C(std::forward<Args>(args)...);
						}
					}

					// ensure the old archetype is still contigious
					const int lastEnt = oldArch->entities.size() - 1;
					if (lastEnt != record.index) {
						for (std::size_t i = 0; i < oldArch->type.size(); ++i) {
							const component_type_t typeID = oldArch->type[i];
							const ComponentBase *const comp = m_componentMap[typeID];
							const std::size_t &compSize = comp->size();

							// shift the empty record to the end of the list
							std::size_t slotOffset = record.index * compSize;
							std::size_t lastOffset = lastEnt * compSize;

							// if we're not the last entity, swap
							if (slotOffset != lastOffset) {
								comp->move(&oldArch->data[i][lastOffset],
										   &oldArch->data[i][slotOffset]);
								comp->destroy(&oldArch->data[i][lastOffset]);
							}
						}

						// fix the position
						oldArch->entities[record.index] = oldArch->entities[lastEnt];
					}
					oldArch->entities.pop_back();
				}

				// register the new data with the new archetype
				newArch->entities.push_back(id);
				record.index = newArch->entities.size() - 1;
				record.archetype = newArch;
				return newComp;
			}

			template<class C>
			void removeComponent(const entity_t &entityId);

			template<class C>
			bool hasComponent(const entity_t &entityId) const {
				const component_type_t componentID = Component<C>::typeID();
				const auto *arch = m_entityArchtypes.at(entityId).archetype;
				if (arch == nullptr) {
					return false;
				}
				return (std::find(arch->type.begin(), arch->type.end(), componentID) !=
					arch->type.end());
			}

			template<class C>
			C *getComponent(const entity_t &entityId) {
				assert(hasComponent<C>(entityId));
				const auto type = Component<C>::typeID();

				const ComponentBase *const newComp = m_componentMap[type];
				const auto record = m_entityArchtypes.at(entityId);
				const auto *arch = record.archetype;

				// JWR: linear search... seems a little suspect, they're ordered after all
				for (std::size_t i = 0; i < arch->type.size(); ++i) {
					if (arch->type[i] == type) {
						return reinterpret_cast<C *>(&(arch->data[i][record.index * newComp->size()]));
					}
				}

				return nullptr;
			}

			template<class C>
			const C *getComponent(const entity_t &entityId) const {
				assert(hasComponent<C>(entityId));
				const auto type = Component<C>::typeID();

				const ComponentBase *const newComp = m_componentMap.at(type);
				const auto record = m_entityArchtypes.at(entityId);
				const auto *arch = record.archetype;

				// JWR: linear search... seems a little suspect, they're ordered after all
				for (std::size_t i = 0; i < arch->type.size(); ++i) {
					if (arch->type[i] == type) {
						return reinterpret_cast<C *>(&(arch->data[i][record.index * newComp->size()]));
					}
				}

				return nullptr;
			}

			template<class... Cs>
			std::vector<entity_t> getEntityWith() const {
				// construct the key
				archToken_t key;
				(key.push_back(Component<Cs>::typeID()), ...);

				// entities
				std::vector<entity_t> entities;
				for (Archetype *arch : m_archetypes) {
					if (std::includes(arch->type.begin(), arch->type.end(), key.begin(), key.end())) {
						if (!arch->entities.empty()) {
							entities.insert(entities.begin(), arch->entities.begin(), arch->entities.end());
						}
					}
				}

				return entities;
			}

		private:
			entitymap_t m_entityArchtypes;
			archetype_t m_archetypes;
			entity_t m_entityIDCounter;
			componentmap_t m_componentMap;
			Archetype *getArchetype(const archToken_t &id);

			inline std::string arch2str(Archetype *arch) {
				std::string str;
				for (const auto &type : arch->type) {
					str += std::to_string(type);
				}
				return str;
			}

			inline std::size_t ensureCapacity(Archetype *arch, const int idx, const ComponentBase *const comp) {
				const std::size_t &compSize = comp->size();
				std::size_t currSize = arch->entities.size() * compSize;
				std::size_t newSize = currSize + compSize;

				std::size_t cap = arch->dataSizes[idx];

				if (newSize > arch->dataSizes[idx]) {
					arch->dataSizes[idx] *= 2;
					arch->dataSizes[idx] += compSize;

					// copy data over
					unsigned char *newData = new unsigned char[arch->dataSizes[idx]];
					for (std::size_t e = 0; e < arch->entities.size(); ++e) {
						const int offset = e * compSize;
						comp->move(&arch->data[idx][offset], &newData[offset]);
						comp->destroy(&arch->data[idx][offset]);
					}

					// free the old data and swap the pointers
					delete[] arch->data[idx];
					arch->data[idx] = newData;
				}

				return currSize;
			}

			inline int getComponentIdx(const Archetype *arch, const component_type_t goal) {
				// JWR could do binary search for speedup
				for (std::size_t i = 0; i < arch->type.size(); ++i) {
					if (arch->type[i] == goal)
						return i;
				}
				return -1;
			}

	};

}

#endif
