#ifndef FGGL_GFX_RENDERER_H
#define FGGL_GFX_RENDERER_H

#include <fggl/data/model.hpp>
#include <fggl/ecs3/ecs.hpp>
#include <fggl/gfx/ogl/backend.hpp>
#include <fggl/gfx/ogl/shader.hpp>

#include "fggl/gfx/ogl4/models.hpp"
#include "fggl/gfx/ogl4/canvas.hpp"

#include <fggl/gfx/paint.hpp>

#include <memory>
#include <vector>

namespace fggl::gfx {

	enum GlRenderType {
		triangles = GL_TRIANGLES,
		triangle_strip = GL_TRIANGLE_STRIP
	};

	struct GlRenderToken {
		constexpr static const char name[] = "RenderToken";
		GLuint vao;
		GLuint buffs[2];
		GLuint idxOffset;
		GLsizei idxSize;
		GLuint pipeline;
		GLuint restartVertex;
		GlRenderType renderType = triangles;
	};

	class GlMeshRenderer {
		public:
			using token_t = GlRenderToken;

			token_t upload(fggl::data::Mesh &mesh);

			void render(ecs3::World &ecs, ecs3::entity_t camera, float dt);

			float total;
	};

	/**
	 * Class responsible for managing the OpenGL context.
	 *
	 * This class manages the OpenGL resources, all OpenGL resources are bound by its lifetime, and it is ultimately
	 * the class responsible for interacting with OpenGL. It is bound by the window it belongs to, and therefore if
	 * that window disappears so does this object.
	 */
	class OpenGL4Backend : public Graphics {
		public:
			explicit OpenGL4Backend(const Window &owner);
			~OpenGL4Backend() override = default;

			// copy bad
			OpenGL4Backend(const OpenGL4Backend&) = delete;
			OpenGL4Backend& operator=(const OpenGL4Backend&) = delete;

			// move (probably) bad
			OpenGL4Backend(OpenGL4Backend&&) = delete;
			OpenGL4Backend&& operator=(OpenGL4Backend&&) = delete;

			/**
			 * Clear the backing buffer.
			 */
			void clear() override;

			/**
			 * Notify the graphics system the framebuffer size has changed.
			 *
			 * @param width the new framebuffer width
			 * @param height the new framebuffer height
			 */
			void resize(int width, int height) override;

			/**
			 * Perform a 2D rendering pass.
			 *
			 * @param paint the objects to paint on screen
			 */
			void draw2D(const Paint &paint) override;

			/**
			 * Perform a 3D rendering pass.
			 *
			 * @param world the world to render
			 */
			void drawScene(ecs3::World& world) override;

			/**
			 * Get the 2D canvas bounds.
			 *
			 * @return
			 */
			inline Bounds canvasBounds() override {
				return m_canvasRenderer->bounds();
			}

		private:
			std::unique_ptr<ogl4::StaticModelRenderer> m_modelRenderer;
			std::unique_ptr<ogl4::CanvasRenderer> m_canvasRenderer;
			std::unique_ptr<ShaderCache> m_cache;
	};

	using OpenGL4 = OpenGL4Backend;

};  // namespace fggl::gfx

#endif
