/*
 * ${license.title}
 * Copyright (C) 2022 ${license.owner}
 * ${license.mailto}
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

//
// uses RAII, with reference to https://www.khronos.org/opengl/wiki/Common_Mistakes#RAII_and_hidden_destructor_calls
//

#ifndef FGGL_GFX_OGL_TYPES_HPP
#define FGGL_GFX_OGL_TYPES_HPP

#include <cstdint>
#include <cassert>
#include <string_view>
#include <iostream>
#include <glm/gtc/type_ptr.hpp>

#include "fggl/gfx/ogl/common.hpp"
#include "fggl/math/types.hpp"

namespace fggl::gfx::ogl {

	// reset values into their 'empty' states
	void deactivate_shader();
	void deactivate_buffer();
	void deactivate_vertex_array();

	constexpr GLuint UNIFORM_NOOP = -1;
	using Location = GLint;

	class Shader {
		private:
			GLuint m_obj = 0;
			void release();

		public:
			Shader() = default;
			inline Shader(GLuint obj) : m_obj(obj) {}

			// copy constructor bad
			Shader(const Shader&) = delete;
			Shader& operator=(const Shader&) = delete;

			// Move ok - if handled properly
			Shader(Shader&& other);
			Shader& operator=(Shader&& other);

			void use() {
				glUseProgram( m_obj );
			}

			inline Location uniform(const std::string_view& name) const {
				return glGetUniformLocation( m_obj, name.data() );
			}

			// primatives
			inline void setUniformF(Location name, GLfloat value) {
				glProgramUniform1f(m_obj, name, value);
			}
			inline void setUniformI(Location name, GLint value) {
				glProgramUniform1i(m_obj, name, value);
			}

			inline void setUniformU(Location name, GLuint value) {
				glProgramUniform1ui(m_obj, name, value);
			}

			// vector versions (float)
			inline void setUniformF(Location name, math::vec2f value) {
				glProgramUniform2f(m_obj, name, value.x, value.y);
			}

			inline void setUniformF(Location name, math::vec3f value) {
				glProgramUniform3f(m_obj, name, value.x, value.y, value.z);
			}

			inline void setUniformF(Location name, math::vec4f value) {
				glProgramUniform4f(m_obj, name, value.x, value.y, value.z, value.w);
			}

			// vector versions (int)
			inline void setUniformI(Location name, math::vec2i value) {
				glProgramUniform2i(m_obj, name, value.x, value.y);
			}

			inline void setUniformI(Location name, math::vec3i value) {
				glProgramUniform3i(m_obj, name, value.x, value.y, value.z);
			}

			inline void setUniformI(Location name, math::vec4i value) {
				glProgramUniform4i(m_obj, name, value.x, value.y, value.z, value.w);
			}

			// matrix versions
			inline void setUniformMtx(Location name, const math::mat2& mtx) {
				glProgramUniformMatrix2fv(m_obj, name, 1, GL_FALSE, glm::value_ptr(mtx));
			}

			void setUniformMtx(Location name, const math::mat3& mtx) {
				glProgramUniformMatrix3fv(m_obj, name, 1, GL_FALSE, glm::value_ptr(mtx));
			}

			void setUniformMtx(Location name, const math::mat4& mtx) {
				glProgramUniformMatrix4fv(m_obj, name, 1, GL_FALSE, glm::value_ptr(mtx));
			}
	};

	enum class Wrapping {
		REPEAT = GL_REPEAT,
		MIRRORED_REPEAT = GL_MIRRORED_REPEAT,
		CLAMP_TO_EDGE = GL_CLAMP_TO_EDGE,
		CLAMP_TO_BORDER = GL_CLAMP_TO_BORDER
	};

	enum class Filtering {
		NEIGHBOUR = GL_NEAREST,
		LINEAR = GL_LINEAR,

		// these are their openGL names, but they are kinda bad, can we do better/alias?
		NEAREST_MIPMAP_NEAREST = GL_NEAREST_MIPMAP_NEAREST,
		LINEAR_MIPMAP_NEAREST = GL_LINEAR_MIPMAP_NEAREST,
		NEAREST_MIPMAP_LINEAR = GL_NEAREST_MIPMAP_LINEAR,
		LINEAR_MIPMAP_LINEAR = GL_LINEAR_MIPMAP_LINEAR,

		MIPMAP_CLOSEST_NEIGHBOUR = GL_NEAREST_MIPMAP_NEAREST,
		MIPMAP_CLOSEST_LINEAR = GL_LINEAR_MIPMAP_NEAREST,
		TWO_MIPMAP_NEIGHBOUR = GL_NEAREST_MIPMAP_LINEAR,
		TWO_MIPMAP_LINEAR = GL_LINEAR_MIPMAP_LINEAR
	};

	enum class TextureType {
		Tex1D = GL_TEXTURE_1D,
		Tex1DArray = GL_TEXTURE_1D_ARRAY,

		Tex2D = GL_TEXTURE_2D,
		Tex2DArray = GL_TEXTURE_2D_ARRAY,
		Tex2DMultisample = GL_TEXTURE_2D_MULTISAMPLE,
		Tex2DMultisampleArray = GL_TEXTURE_2D_MULTISAMPLE_ARRAY,
		Tex3D = GL_TEXTURE_3D,

		Rectangle = GL_TEXTURE_RECTANGLE,
		CubeMap = GL_TEXTURE_CUBE_MAP,
		CubeMapArray = GL_TEXTURE_CUBE_MAP_ARRAY,
		Buffer = GL_TEXTURE_BUFFER,

	};

	enum class PixelFormat {
			UNSIGNED_BYTE = GL_UNSIGNED_BYTE,
			BYTE,
			UNSIGNED_SHORT,
			SHORT,
			UNSIGNED_INT,
			INT,
			HALF_FLOAT,
			FLOAT,

			UNSIGNED_BYTE_3_3_2,
			UNSIGNED_BYTE_2_3_3_REV,
			UNSIGNED_SHORT_5_6_5,
			UNSIGNED_SHORT_5_6_5_REV,
			UNSINGED_SHORT_4_4_4_4,
			UNSIGNED_SHORT_4_4_4_4_REV,
			UNSIGNED_SHORT_5_5_5_1,
			UNSIGNED_SHORT_1_5_5_5_REV,
			UNSIGNED_INT_8_8_8_8,
			UNSIGNED_INT_8_8_8_8_REV,
			UNSIGNED_INT_10_10_10_10_2,
			UNSIGNED_INT_10_10_10_10_2_REV,
	};

	enum class InternalImageFormat {
			DepthComponent = GL_DEPTH_COMPONENT,
			DepthStencil = GL_DEPTH_STENCIL,
			Red = GL_RED,
			RedGreen = GL_RG,
			RedGreenBlue = GL_RGB,
			RedGreenBlueAlpha = GL_RGBA
	};

	enum class ImageFormat {
			R = GL_RED,
			RG,
			RGB,
			RGBA,
			R_INT,
			RG_INT,
			RGB_INT,
			RGBA_INT,
			BGR,
			BGRA,
			BGR_INT,
			BGRA_INT,
			STENTICL_INDEX,
			DEPTH_COMPONENT,
			DEPTH_STENCIL
	};

	struct Image {
		PixelFormat type;
		ImageFormat format;
		math::vec2i size;
		void* data;
	};

	class Texture {
		public:
			inline explicit Texture(TextureType type) : m_type(type) {
				glGenTextures(1, &m_obj);
			}
			~Texture() {
				glDeleteTextures(1, &m_obj);
			}

			void setup(InternalImageFormat iFmt, math::vec2i size) {
				//bind();
				glBindTexture( (GLenum)m_type, m_obj );
				if ( iFmt == InternalImageFormat::DepthComponent ) {
					glTexImage2D((GLenum)
					m_type, 0, (GLint) iFmt, size.x, size.y, 0, GL_DEPTH_COMPONENT, GL_UNSIGNED_BYTE, nullptr);
				} else {
					glTexImage2D((GLenum)
					m_type, 0, (GLint) iFmt, size.x, size.y, 0, GL_RGB, GL_UNSIGNED_BYTE, nullptr);
				}
			}

			void setData(InternalImageFormat iFmt, Image& image) {
				//bind();
				glBindTexture( (GLenum)m_type, m_obj );
				glTexImage2D( (GLenum)m_type, 0, (GLint)iFmt, image.size.x, image.size.y, 0, (GLenum)image.format, (GLenum)image.type, image.data);
			}

			void setDataPart(math::vec2i offset, Image& image) {
				glTexImage2D( (GLenum)m_type, 0, offset.x, offset.y, image.size.x, image.size.y, (GLenum)image.format, (GLenum)image.type, image.data);
			}

			void wrapMode(Wrapping wrap);
			void filterModeMagnify(Filtering filter);
			void filterModeMinify(Filtering filter);

			/**
			 * Bind this texture to a texture unit.
			 *
			 * @param textureUnit the texture unit to bind to
			 */
			inline void bind(unsigned int textureUnit) {
				assert( 0 <= textureUnit && textureUnit < GL_MAX_COMBINED_TEXTURE_IMAGE_UNITS );
				glActiveTexture( GL_TEXTURE0 + textureUnit );
				glBindTexture( (GLenum) m_type,  m_obj );
			}

		private:
			GLuint m_obj = 0;
			TextureType m_type;

			void release();
	};

	enum class BuffAttrF {
		HALF_FLOAT = GL_HALF_FLOAT,
		FLOAT = GL_FLOAT,
		DOUBLE = GL_DOUBLE,
		FIXED = GL_FIXED
	};

	enum class BuffAttrI{
		BYTE = GL_BYTE,
		UBYTE = GL_UNSIGNED_BYTE,
		SHORT = GL_SHORT,
		USHORT = GL_UNSIGNED_SHORT,
		INT = GL_INT,
		UINT = GL_UNSIGNED_INT,

		// these ones are oddly spesific...
		INT_PACKED = GL_INT_2_10_10_10_REV,
		UINT_PACKED = GL_UNSIGNED_INT_2_10_10_10_REV,
		UINT_PACKED_F = GL_UNSIGNED_INT_10F_11F_11F_REV
	};

	enum class Primative {
		POINT = GL_POINT,
		LINE = GL_LINES,
		LINE_STRIP = GL_LINE_STRIP,
		LINE_LOOP = GL_LINE_LOOP,
		TRIANGLE = GL_TRIANGLES,
		TRIANGLE_STRIP = GL_TRIANGLE_STRIP,
		TRIANGLE_FAN = GL_TRIANGLE_FAN
	};

	enum class BufType {
		ARRAY = GL_ARRAY_BUFFER,
		ELEMENT_ARRAY = GL_ELEMENT_ARRAY_BUFFER,
		COPY_READ = GL_COPY_READ_BUFFER,
		COPY_WRITE = GL_COPY_WRITE_BUFFER,
		PIXEL_UNPACK = GL_PIXEL_UNPACK_BUFFER,
		PIXEL_PACK = GL_PIXEL_PACK_BUFFER,
		QUERY = GL_QUERY_BUFFER,
		TEXTURE = GL_TEXTURE_BUFFER,
		TRANSFORM_FEEDBACK = GL_TRANSFORM_FEEDBACK_BUFFER,
		UNIFORM = GL_UNIFORM_BUFFER,
		DRAW_INDRECT = GL_DRAW_INDIRECT_BUFFER,
		ATOMIC_COUNTER = GL_ATOMIC_COUNTER_BUFFER,
		DISPATCH_INDIRECT = GL_DISPATCH_INDIRECT_BUFFER,
		SHADER_STORAGE = GL_SHADER_STORAGE_BUFFER
	};

	enum class BufUsage {
		STATIC_DRAW = GL_STATIC_DRAW,
		STATIC_READ = GL_STATIC_READ,
		STATIC_COPY = GL_STATIC_COPY,
		DYNAMIC_DRAW = GL_DYNAMIC_DRAW,
		DYNAMIC_READ = GL_DYNAMIC_READ,
		DYNAMIC_COPY = GL_DYNAMIC_COPY,
		STREAM_DRAW = GL_STREAM_DRAW,
		STREAM_READ = GL_STREAM_READ,
		STREAM_COPY = GL_STREAM_COPY,
	};

	template<BufType T>
	class Buffer {
		private:
			GLuint m_obj = 0;
			GLsizeiptr m_capacity = 0;

			/**
			 * Free an underlying buffer.
			 *
			 * If this buffer object is already empty, this is a no-op (and the OpenGL spec guarntees that freeing buffer
			 * 0 has no effect.
			 */
			void release() {
				glDeleteBuffers(1, &m_obj);
				m_obj = 0;
			}

		public:
			/**
			 * Create a new (unfilled) buffer object.
			 */
			Buffer() {
				glGenBuffers(1, &m_obj);
			}

			~Buffer() {
				release();
			}

			// copy constructor bad
			Buffer(const Buffer&) = delete;
			Buffer& operator=(const Buffer&) = delete;

			Buffer(Buffer&& other) : m_obj(other.m_obj), m_capacity(other.m_capacity) {
				other.obj_ = 0;
				other.m_capacity = 0;
			}

			Buffer& operator=(Buffer&& other) {
				if ( this != &other) {
					release();
					std::swap( m_obj, other.m_obj );
					std::swap( m_capacity, other.m_capacity );
				}
			}

			void bind() const {
				assert( m_obj != 0 );
				glBindBuffer( (GLenum)T, m_obj );
			}

			inline bool isValid() const {
				return m_obj != 0;
			};

			void write(GLsizeiptr size, const GLvoid* data, BufUsage usage) {
				bind();
				glBufferData( (GLenum)T, size, data, (GLenum)usage);
			}

			void update(GLintptr offset, GLsizeiptr size, const void* data) {

			}

			template<typename D>
			void replace(std::size_t size, D* data) {
				bind();
				GLsizeiptr sizePtr = size * sizeof(D);
				if ( sizePtr > m_capacity) {
					glBufferData( (GLenum)T, sizePtr, data, GL_STREAM_DRAW );
					m_capacity = sizePtr;
				} else {
					glBufferSubData((GLenum)T, 0, sizePtr, data);
				}
				glBindBuffer( (GLenum)T, 0 );
			}

	};

	// common buffer types
	using ArrayBuffer = Buffer<BufType::ARRAY>;
	using ElementBuffer = Buffer<BufType::ELEMENT_ARRAY>;

	struct AttributeF {
		BuffAttrF attrType;
		GLint elmCount;
		GLsizei stride;
		std::size_t offset;
	};

	struct AttributeI {
		BuffAttrI attrType;
		GLint elmCount;
		GLsizei stride;
		std::size_t offset;
	};

	// type intrincs to make interface nicer
	template<typename T>
	struct attr_type{
		const static BuffAttrF attr;
		const static GLint size;
	};

	template<typename V, typename T>
	AttributeF attribute(std::size_t offset) {
		return AttributeF{
			attr_type<T>::attr,
			attr_type<T>::size,
			sizeof(V),
			offset
		};
	}

	/**
	 * Represents an OpenGL vertex array object (VAO).
	 *
	 * This contains most information about an OpenGL state for supplying vertex data to the GPU. There is some basic
	 * debug checking to stop silly stuff happening, but this will be disabled in release mode.
	 */
	class VertexArray {
		private:
			GLuint m_obj = 0;
			void release();

		public:
			VertexArray();
			~VertexArray();

			// copy constructors bad
			VertexArray(const VertexArray&) = delete;
			VertexArray& operator=(const VertexArray) = delete;

			// move constructors might be ok
			VertexArray(VertexArray&& other) noexcept;
			VertexArray& operator=(VertexArray&& other);

			inline void bind() const {
				assert( m_obj != 0);
				glBindVertexArray( m_obj );
			}

			void setAttribute(const ArrayBuffer& buffer, GLuint idx, AttributeF& attr);
			void setAttribute(const ArrayBuffer& buffer, GLuint idx, AttributeI& attr, bool normalized);
			void setAttributeI(const ArrayBuffer& buffer, GLuint idx, AttributeI& attr);

			void drawElements(const ElementBuffer& buff, Primative drawType, std::size_t size);
			void draw(Primative drawType, int first, std::size_t count);
	};

	// paranoid functions
	void bind_vertex_array(GLuint& marker);
	void unbind_vertex_array(GLuint& marker);

	template<BufType T>
	void bind_buffer(GLuint* marker, const Buffer<T>& buff) {
		#ifdef FGGL_GL_PARANOID
		assert( marker != nullptr );
			glGetIntegerv( (GLenum)T, (GLint*) marker );
		#endif
		buff.bind();
	}

	template<BufType T>
	void unbind_buffer(GLuint* marker, const Buffer<T>& buff) {
		#ifdef GL_FGGL_PARANOID
		assert( marker != nullptr );
			glBindVertexArray(marker);
		#endif
	}

} // namespace fggl::gfx::ogl

#endif //FGGL_GFX_OGL_TYPES_HPP
