/*
 * ${license.title}
 * Copyright (C) 2022 ${license.owner}
 * ${license.mailto}
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU Lesser General Public
 * License as published by the Free Software Foundation; either
 * version 3 of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
 * Lesser General Public License for more details.
 *
 * You should have received a copy of the GNU Lesser General Public License
 * along with this program; if not, write to the Free Software Foundation,
 * Inc., 51 Franklin Street, Fifth Floor, Boston, MA  02110-1301, USA.
 */

//
// Created by webpigeon on 21/04/22.
//

#ifndef FGGL_GUI_FONTS_HPP
#define FGGL_GUI_FONTS_HPP

#include <memory>
#include <map>
#include <string>

#include "fggl/math/types.hpp"
#include "fggl/data/storage.hpp"
#include "fggl/util/service.h"

#include <ft2build.h>
#include FT_FREETYPE_H

namespace fggl::gui {

	struct GlyphMetrics {
		math::vec2 size;
		math::vec2 bearing;
		long advance;
	};

	class FontFace {
		public:
			explicit FontFace(FT_Face face);
			~FontFace();

			FontFace(const FontFace&) = delete;
			FontFace(FontFace&&) = delete;
			FontFace& operator=(const FontFace&) = delete;
			FontFace& operator=(FontFace&&) = delete;

			inline GlyphMetrics& metrics(char letter){
				auto itr = m_metrics.find(letter);
				if ( itr == m_metrics.end() ) {
					return populateMetrics(letter);
				}
				return itr->second;
			}

			math::vec2 stringSize(const std::string& text);
			void texture(char letter, int& width, int& height, void** buff);

		private:
			FT_Face m_face;
			std::map<char, GlyphMetrics> m_metrics;

			GlyphMetrics& populateMetrics(char letter);
	};

	class FontLibrary {
		public:
			FontLibrary();
			~FontLibrary();

			// copy and moving not needed
			FontLibrary(const FontLibrary&) = delete;
			FontLibrary(FontLibrary&&) = delete;
			FontLibrary& operator=(const FontLibrary&) = delete;
			FontLibrary& operator=(FontLibrary&&) = delete;

			inline std::shared_ptr<FontFace> getFont(const std::string& name) {
				auto fontItr = m_cache.find(name);
				if ( fontItr != m_cache.end() ) {
					return fontItr->second;
				}

				// need to load the font...
				auto storage = util::ServiceLocator::instance().get<data::Storage>();
				auto path = storage->resolvePath(data::StorageType::Data, name);

				FT_Face face;
				if ( FT_New_Face(m_context, path.c_str(), 0, &face) ) {
					return nullptr;
				}
				FT_Set_Pixel_Sizes(face, 0, 18);

				// create the new font object
				auto ptr = std::make_shared<FontFace>(face);
				m_cache[name] = ptr;
				return ptr;
			}

		private:
			FT_Library m_context;
			std::map<const std::string, std::shared_ptr<FontFace>> m_cache;
	};



} // nmespace fggl::gui

#endif //FGGL_GUI_FONTS_HPP
