#ifndef FGGL_GUI_WIDGET_H
#define FGGL_GUI_WIDGET_H

#include <fggl/math/types.hpp>
#include <fggl/gfx/paint.hpp>

#include <memory>

namespace fggl::gui {

	constexpr float RGB_MAX_VAL = 255.0F;
	constexpr math::vec3 rgbToNormal(char red, char green, char blue) {
		return {red / RGB_MAX_VAL, green / RGB_MAX_VAL, blue / RGB_MAX_VAL};
	}

	/**
	 * Convert a Hue-saturation-value colour to RGB
	 *
	 * @param hue hue component, assumed 0 - 360
	 * @param saturation assumed 0 - 1
	 * @param value assumed 0 - 1
	 * @return RGB, in the range 0,1 for each component
	 */
	constexpr math::vec3 HSVtoNormal(float hue, float saturation, float value) {
		assert( 0 < hue && hue <= 360);
		assert( 0 < saturation && saturation <= 1);
		assert( 0 < value && value <= 1);

		const float chroma = value * saturation;
		const float x = chroma * (1 - fabs( fmod(hue / 60.0F, 2) - 1) );
		math::vec3 tmp{0,0,0};
		if ( 0 <= hue && hue < 60 ) {
			tmp = {chroma,x,0};
		} else if ( 60 <= hue && hue < 120 ) {
			tmp = {x, chroma, 0};
		} else if (120 <= hue && hue < 180 ) {
			tmp = {0, chroma, x};
		} else if ( 180 <= hue && hue < 240 ) {
			tmp = {0, x, chroma};
		} else if ( 240 <= hue && hue < 300) {
			tmp = {x, 0, chroma};
		} else if ( 300 <= hue && hue < 360) {
			tmp = {chroma, 0, x};
		}
		return tmp + (value - chroma);
	}

	void draw_box( gfx::Path2D& path, math::vec2 topLeft, math::vec2 bottomRight);
	void draw_progress( gfx::Path2D& path, math::vec2 topLeft, math::vec2 size, float value);
	void draw_slider( gfx::Path2D& path, math::vec2 topLeft, math::vec2 size, float value);
	void draw_button( gfx::Path2D& path, math::vec2 topLeft, math::vec2 size, bool active, bool pressed);

	struct Bounds2D {
		math::vec2 topLeft;
		math::vec2 size;

		Bounds2D() = default;
		inline Bounds2D(math::vec2 pos, math::vec2 a_size) : topLeft(pos), size(a_size) {}

		inline bool contains(math::vec2 point) {
			return ! ( (point.x > topLeft.x + size.x) ||
			           (point.x < topLeft.x ) ||
					   (point.y > topLeft.y + size.y) ||
				       (point.y < topLeft.y)
					   );
		}
	};

	class Widget {
		public:
			Widget() = default;
			inline Widget(math::vec2 pos, math::vec2 size) : m_bounds(pos, size) {}

			virtual ~Widget() = default;

			inline math::vec2 topLeft() const {
				return m_bounds.topLeft;
			};

			inline math::vec2 bottomRight() const {
				return m_bounds.topLeft + m_bounds.size;
			}

			inline void size(math::vec2 pos, math::vec2 size) {
				m_bounds.topLeft = pos;
				m_bounds.size = size;
			}

			inline math::vec2 size() const {
				return m_bounds.size;
			}

			virtual inline bool contains(const math::vec2 &point){
				return m_bounds.contains(point);
			};

			virtual inline Widget *getChildAt(const math::vec2 &point) {
				if ( !contains(point) ) {
					return nullptr;
				}
				return this;
			}

			virtual void render(gfx::Paint &paint) = 0;
			inline virtual void activate() {};

			inline virtual void onEnter() {}
			inline virtual void onExit() {}

		private:
			Bounds2D m_bounds;
	};


}; //namespace fggl::gui

#endif
