#ifndef FGGL_INPUT_MOUSE_H
#define FGGL_INPUT_MOUSE_H

#include <string>
#include <array>
#include <bitset>

namespace fggl::input {

	enum class MouseButton {
		LEFT,
		MIDDLE,
		RIGHT,
		EXTRA_1,
		EXTRA_2,
		EXTRA_3,
		EXTRA_4,
		EXTRA_5
	};

	enum class MouseAxis {
		X,
		Y,
		SCROLL_X,
		SCROLL_Y
	};

	struct MouseButtonRecord {
		MouseButton id;
		const char name[15];
	};

	struct MouseAxisRecord {
		MouseAxis id;
		const char name[15];
	};

	constexpr std::array<MouseButtonRecord, 8> MouseButtons = {{
																   {MouseButton::LEFT, "Left"},
																   {MouseButton::MIDDLE, "Middle"},
																   {MouseButton::RIGHT, "Right"},
																   {MouseButton::EXTRA_1, "Extra 1"},
																   {MouseButton::EXTRA_2, "Extra 2"},
																   {MouseButton::EXTRA_3, "Extra 3"},
																   {MouseButton::EXTRA_4, "Extra 4"},
																   {MouseButton::EXTRA_5, "Extra 5"},
															   }};

	constexpr std::array<MouseAxisRecord, 4> MouseAxes = {{
															  {MouseAxis::X, "Left/Right"},
															  {MouseAxis::Y, "Up/Down"},
															  {MouseAxis::SCROLL_X, "Scroll X"},
															  {MouseAxis::SCROLL_Y, "Scroll Y"}
														  }};

	struct MouseState {
		float axis[MouseAxes.size()];
		std::bitset<MouseButtons.size()> buttons;

		void operator=(const MouseState &rhs);
	};

	class MouseInput {
		public:
			MouseInput() = default;
			MouseInput(const MouseInput &rhs) = delete;
			void operator=(const MouseInput &rhs) = delete;

			inline void frame(float dt) {
				m_prev = m_curr;
			}

			inline void axis(MouseAxis axis, float value) {
				m_curr.axis[(int) axis] = value;
			}

			inline float axis(MouseAxis axis) const {
				return m_curr.axis[(int) axis];
			}

			inline float axisDelta(MouseAxis axis) const {
				return m_curr.axis[(int) axis] - m_prev.axis[(int) axis];
			}

			inline void button(MouseButton btn, bool state) {
				m_curr.buttons[(int) btn] = state;
			}

			inline bool down(MouseButton btn) const {
				return m_curr.buttons[(int) btn];
			}

			inline bool downPrev(MouseButton btn) const {
				return m_prev.buttons[(int) btn];
			}

			inline bool pressed(MouseButton btn) const {
				return m_curr.buttons[(int) btn] && !m_prev.buttons[(int) btn];
			}

			inline bool released(MouseButton btn) const {
				return !m_curr.buttons[(int) btn] && m_prev.buttons[(int) btn];
			}

		private:
			MouseState m_curr;
			MouseState m_prev;
	};

};

#endif
