#ifndef FGGL_MATH_TYPES_H
#define FGGL_MATH_TYPES_H

#include <tuple>

#include <glm/glm.hpp>
#include <glm/ext/matrix_transform.hpp>
#include <glm/gtc/quaternion.hpp>
#include <glm/gtx/quaternion.hpp>

#ifndef M_PI
    #define M_PI 3.14159265358979323846
#endif

#ifndef M_PI_2
    #define M_PI_2 1.57079632679489661923
#endif

namespace fggl::math {

	constexpr float PI = M_PI;
	constexpr float HALF_PI = M_PI_2;

	// math types (aliased for ease of use)
	using vec4 = glm::vec4;
	using vec4f = glm::vec4;
	using vec4i = glm::ivec4;
	using vec4ui = glm::ivec4;

	using vec3 = glm::vec3;
	using vec3i = glm::ivec3;
	using vec3f = glm::vec3;
	using vec3ui = glm::ivec3;

	using vec2 = glm::vec2;
	using vec2f = glm::vec2;
	using vec2i = glm::ivec2;
	using vec2ui = glm::ivec2;

	using mat2 = glm::mat2;
	using mat3 = glm::mat3;
	using mat4 = glm::mat4;
	using quat = glm::quat;

	// fastFloor from OpenSimplex2
	inline int fastFloor(double x) {
		int xi = (int) x;
		return x < xi ? xi - 1 : xi;
	}

	inline float rescale_norm(float value, float min, float max) {
		return (value - min) / (max - min);
	}

	inline float rescale_norm(float value, float min, float max, float newMin, float newMax) {
		return newMin + ((value - min) * (newMax - newMin)) / (max - min);
	}

	inline float rescale_ndc(float value, float newMin, float newMax){
		return rescale_norm(value, -1, 1, newMin, newMax);
	}

	inline float rescale_01(float value, float newMin, float newMax){
		return rescale_norm(value, 0, 1, newMin, newMax);
	}

	inline float recale_mean(float value, float avg, float max, float min) {
		return (value - avg) / (max - min);
	}


	// reference vectors
	constexpr vec3f UP{0.0f, 1.0f, 0.0f};
	constexpr vec3f FORWARD{1.0f, 0.0f, 0.0f};
	constexpr vec3f RIGHT{0.0f, 0.0f, 1.0f};

	inline glm::mat4 modelMatrix(const vec3 offset, const quat rotation) {
		return glm::translate(glm::mat4(1.0f), offset) * glm::toMat4(rotation);
	}

	inline glm::mat4 modelMatrix(glm::vec3 offset, glm::vec3 eulerAngles) {
		return modelMatrix(offset, glm::quat(eulerAngles));
	}

	struct Transform {
			constexpr static const char name[] = "Transform";

			Transform() : m_local(1.0f), m_origin(0.0f), m_model(1.0f), m_rotation() {
			}

			// local reference vectors
			[[nodiscard]]
			inline vec3 up() const {
				return vec4(UP, 1.0) * m_local;
			}

			[[nodiscard]]
			inline vec3 forward() const {
				return vec4(FORWARD, 1.0) * m_local;
			}

			[[nodiscard]]
			inline vec3 right() const {
				return vec4(RIGHT, 1.0) * m_local;
			}

			inline void translate(const vec3 change) {
				m_origin += change;
				update();
			}

			inline void origin(const vec3 pos) {
				m_origin = pos;
				update();
			}

			[[nodiscard]]
			inline vec3 origin() const {
				return m_origin;
			}

			inline void euler(vec3 angles) {
				m_rotation = quat(angles);
				update();
			}

			[[nodiscard]]
			inline glm::vec3 euler() const {
				return glm::eulerAngles(m_rotation);
			}

			inline mat4 model() const {
				mat4 tmp(1.0f);
				tmp = glm::translate(tmp, m_origin);
				tmp = tmp * glm::toMat4(m_rotation);

				return tmp;
			}

		private:
			mat4 m_local; // us -> parent
			mat4 m_model; // us -> world
			vec3 m_origin;
			quat m_rotation;

			inline void update() {
				mat4 t(1.0f);
				t *= glm::toMat4(m_rotation);
				t = glm::translate(t, m_origin);
				m_local = t;
			}
	};

}


// feels a bit strange to be doing this...
namespace glm {
	inline bool operator<(const vec3 &lhs, const vec3 &rhs) {
		return std::tie(lhs.x, lhs.y, lhs.z)
			< std::tie(rhs.x, rhs.y, rhs.z);
	}
}

#endif
